#!/usr/bin/env python
# encoding: utf-8
"""
correct_snps_from_pileup.py

Created by Rod Docking on 2010-04-27.
Copyright (c) 2010 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import re
import string
from optparse import OptionParser
from lib.common_utils import build_header_dict_from_fai, long_seq_to_lines_of_fifty
from lib.pileup_record import Pileup_record
from lib.vcf_record import VCF_record

def parse_options():
    """Parse options provided on the command line"""
    
    usage = "usage: %prog [options] snps.pileup.txt reference.fa output.fa edits_list.txt"
    parser = OptionParser(usage)
    
    parser.add_option("-o", "--hom",
        action="store_true", dest="hom", 
        help="correct homozygous SNPs")
    parser.add_option("-e", "--het",
        action="store_true", dest="het", 
        help="correct heterozygous SNPs to the majority base")
    parser.add_option("-i", "--indel",
        action="store_true", dest="indel", 
        help="correct indels")
    parser.add_option("-s", "--switch",
        action="store_true", dest="switch", 
        help="switch heterozygous SNPs to the alternate base")
    
    (options, args) = parser.parse_args()
    pathed_args = [os.path.abspath(arg) for arg in args]
        
    if not (len(args) == 4):
        parser.error("incorrect number of arguments")
    
    return options, pathed_args


def parse_pileup(pileup_file, opt):
    """Parse the pileup-style output to return a dictionary of all the fixes that need to be made"""

    pileup_in = open(pileup_file,'r')
    correction_dict = {}
    heterozygous_counter, homozygous_counter, indel_counter, skipped_indel_counter, keep_ref_counter = 0, 0, 0, 0, 0
    
    for line in pileup_in:
        
        p = Pileup_record(line)
        correction_string = ''
                
        # If the called base is ACGT, use that, otherwise determine the majority base
        if p.is_homozygous() and opt.hom:
            homozygous_counter += 1
            correction_string = "{0} {1} {2}".format(p.pos, p.ref_base, p.con_base)

        elif p.is_heterozygous() and opt.het: 
            new_base = resolve_ambiguous_base(p)
            if new_base == p.ref_base:
                keep_ref_counter += 1
            else:
                heterozygous_counter += 1
                correction_string = "{0} {1} {2}".format(p.pos, p.ref_base, new_base)

        elif p.is_indel() and opt.indel:
            indel_to_correct = parse_indel(p)
            if indel_to_correct:
                indel_counter += 1
                correction_string = "{0} {1} {2}".format(p.pos, p.ref_base, p.indel_allele_one)
            else:
                skipped_indel_counter += 1
        
        if correction_string:
        
            if not correction_dict.has_key(p.rname):
                correction_dict[p.rname] = []
                correction_dict[p.rname].append(correction_string)
            else:
                correction_dict[p.rname].append(correction_string)
    
    pileup_in.close()

    print "Homozygous SNPs:", homozygous_counter
    print "Heterozygous SNPs: Altering {0}, Leaving {1}".format(heterozygous_counter, keep_ref_counter)
    print "Indels: Altering {0}, Leaving {1}".format(indel_counter, skipped_indel_counter)
    
    return correction_dict

def parse_VCF(VCF_file, opt):
    """Parse VCF input to find SNPs and indels to correct"""
    
    correction_dict = {}
    heterozygous_counter, homozygous_counter, indel_counter, skipped_indel_counter, keep_ref_counter = 0, 0, 0, 0, 0
    
    for line in open(VCF_file, 'r'):
        
        V = VCF_record(line)
        correction_string = ''
        
        if V.header:
            continue
                
        if V.is_SNP():
            
            if V.is_homozygous() and opt.hom:
                homozygous_counter += 1
                correction_string = "{0} {1} {2}".format(V.pos, V.ref, V.alt)
            
            elif V.is_heterozygous() and opt.switch:
                heterozygous_counter += 1
                correction_string = "{0} {1} {2}".format(V.pos, V.ref, V.alt)
                
            elif V.is_heterozygous() and opt.het:     
                heterozygous_counter += 1
            
        elif V.is_indel():
            
            if opt.indel:
            
                if V.is_homozygous():
                    indel_counter += 1
                    correction_string = "{0} {1} {2}".format(V.pos, V.ref, V.alt)
            
                elif V.is_heterozygous():
                    skipped_indel_counter += 1
            else:
                continue
            
        else:
            print V.raw
            print "Not a SNP or an indel???"
            sys.exit()        

        # If we've found a correction to make, add it to the dictionary
        if correction_string:

            if not correction_dict.has_key(V.chrom):
                correction_dict[V.chrom] = []
                correction_dict[V.chrom].append(correction_string)
            else:
                correction_dict[V.chrom].append(correction_string)

    print "Correcting {0} homozygous positions, {1} heterozygous positions, and {2} indel positions".format(
        homozygous_counter, heterozygous_counter, indel_counter)
    print "SCRUB:\t{0}\t{1}".format(homozygous_counter, indel_counter)

    return correction_dict

def correct_sequence(contig_to_correct, seq_to_correct, header_dict, correction_dict):
    """Correct SNPs in the passed sequence"""

    #print "\nCorrecting: {0}\n{1}".format(contig_to_correct, seq_to_correct)
    #print "Seq length: {0}".format(len(seq_to_correct))
    
    # Check that the length matches what we expect
    #print "Expected: {0} Found: {1}".format(header_dict[contig_to_correct], len(seq_to_correct))
    if not ( len(seq_to_correct) == int(header_dict[contig_to_correct]) ):
        print "Error! The seq length we found doesn't match the header dictionary!"
        sys.exit()
    
    # Need to keep track of the offset that we introduce by adding/removing bases
    #  Initially set to -1: string index is 0-based, pos coordinates from pileup are 1-based
    offset = -1
    
    # For each base to correct:
    for base_to_correct in correction_dict[contig_to_correct]:
        cols = base_to_correct.split()
        position, ref_base, called_base = int(cols[0]), cols[1], cols[2]
        
        # For cases where there are multiple alternate alleles
        if "," in called_base:
            called_base = called_base[0]
        
        # Split SNP-correcting and indel-correcting logic:
        if len(called_base) == 1 and len(ref_base) == 1:
            # print "Ref: {0} Call: {1} Position: {2} Base at position: {3}".format(
            #     ref_base, called_base, position, seq_to_correct[position-1])
                
            # Check that it currently has the expected reference... note that the string index is 0-based, the position is 1-based
            try:
                assert(ref_base.upper() == seq_to_correct[position+offset].upper())
            except AssertionError:    
                print "Error! The reference at the indicated position doesn't match the expected reference from the pileup file!"
                print contig_to_correct
                print seq_to_correct
                print correction_dict[contig_to_correct]
                print base_to_correct
                sys.exit()
            
            # Return a new string with the base corrected
            corrected_seq = seq_to_correct[:position + offset] + called_base + seq_to_correct[position + offset + 1:]
        
            # Check the the length hasn't changed
            if not(len(corrected_seq) == len(seq_to_correct)):
                print "Error! Old and new sequences are not the same length!"
                sys.exit()
        
            # Update for the next round
            seq_to_correct = corrected_seq
        
            # Check that the new string has the right base:    
            try:
                assert(called_base.upper() == seq_to_correct[position+offset].upper())
            except AssertionError:    
                print "Error! The reference at the indicated position doesn't match the expected reference from the pileup file!"
                print contig_to_correct
                print seq_to_correct
                print correction_dict[contig_to_correct]
                print base_to_correct
                sys.exit()
        else:
            
            #print "Correcting an indel"
            
            # Pileup indels and VCF indels will have slightly different formats...
            # Pileup-specific version
            if called_base[0] in "+-":
                sign, seq = called_base[0], called_base[1:]
            
                if sign == '+':

                    # Correct the sequence by inserting the appropriate bases
                    corrected_seq = seq_to_correct[:(position+offset+1)] + seq + seq_to_correct[(position+offset+1):]    

                    # Check that the new sequence is of the appropriate length
                    assert( len(corrected_seq) == len(seq_to_correct)+len(seq) )

                    # Update the sequence and offset                
                    seq_to_correct = corrected_seq
                    offset += len(seq)

                elif sign == '-':

                    # Check that the sequence we deleted matches what we expected to find
                    seq_to_delete = seq_to_correct[(position+offset+1):(position+offset+1+len(ref_base))]
                    assert(seq_to_delete == seq)

                    # Correct the sequence by taking all but the bases to delete
                    # Have to add 1 here: the pileup column is a pad column, so delete the next few bases                
                    corrected_seq = seq_to_correct[:(position+offset+1)] + seq_to_correct[(position+offset+1+len(seq)):]

                    # Check that the new sequence is the appropriate length
                    assert( len(corrected_seq) == len(seq_to_correct)-len(seq) )

                    # Update the sequence and offset                
                    seq_to_correct = corrected_seq
                    offset -= len(seq)
            
            else:
                #print "Ref:", ref_base, "Called:", called_base, "Position:", position, "Current offset:", offset
                
                # Find the ref base in the sequence
                ref_slice = seq_to_correct[position+offset:position+offset+len(ref_base)]
                #print ref_slice
                assert(ref_slice == ref_base)
                
                #print seq_to_correct[:(position+offset)], " + ", called_base, " + ", seq_to_correct[position+offset+len(ref_base):]
                
                corrected_seq = seq_to_correct[:(position+offset)] + called_base + seq_to_correct[position+offset+len(ref_base):]
                assert(len(corrected_seq) == ( len(seq_to_correct) + len(called_base) - len(ref_base) ) )
                                
                seq_to_correct = corrected_seq
                offset += (len(called_base) - len(ref_base))
                #print "New offset:", offset
                #print seq_to_correct

    #print seq_to_correct            
    return seq_to_correct

def correct_assembly_file(header_dict, correction_dict, fasta_file, output_file, edits_list):
    """Parse the assembly fasta file, returning new contigs"""
    
    fasta_in = open(fasta_file,'r')
    fasta_out = open(output_file, 'w')
    edit_handle = open(edits_list, 'a')
    
    fix_this_contig = False
    contig_to_correct, seq_to_correct = '', ''
    
    for line in fasta_in:
        
        # For header lines, check if they're present in the correction dictionary
        if ">" in line:
            header_cols = line.split()
            fasta_ref_name = header_cols[0].lstrip('>')
            
            # If we've reached a new header and we've collected a sequence to correct, correct it
            if contig_to_correct and seq_to_correct:
                corrected_seq = correct_sequence(contig_to_correct, seq_to_correct, header_dict, correction_dict)                      
                corrected_seq_split = long_seq_to_lines_of_fifty(corrected_seq)
                fasta_out.write(corrected_seq_split)
                    
            # Quickly check that we have this contig in our reference
            if not header_dict.has_key(fasta_ref_name):    
                print "Error! Sequence missing from header dictionary!"
                sys.exit()
                
            # If we need to correct this sequence, start collecting it    
            if correction_dict.has_key(fasta_ref_name):    
                #print fasta_ref_name, header_dict[fasta_ref_name], correction_dict[fasta_ref_name]
                fix_this_contig = True
                contig_to_correct = fasta_ref_name
                edit_handle.write("{0}\n".format(fasta_ref_name))
                seq_to_correct = ''
            
            # Otherwise, reset variables and skip this one
            else:
                fix_this_contig = False    
                contig_to_correct, seq_to_correct = '', ''
        
            # Write all headers unmodified
            fasta_out.write(line)
            
        # For sequence lines, either write them directly, or collect them for replacement    
        else:
            
            if fix_this_contig == True:
                seq_to_correct = seq_to_correct + line.rstrip()
            else:
                fasta_out.write(line)

    if contig_to_correct and seq_to_correct:
        corrected_seq = correct_sequence(contig_to_correct, seq_to_correct, header_dict, correction_dict)                      
        corrected_seq_split = long_seq_to_lines_of_fifty(corrected_seq)
        fasta_out.write(corrected_seq_split)
            
    fasta_in.close()
    fasta_out.close()
    edit_handle.close()
    
    return    

def resolve_ambiguous_base(p):
    """From a pileup line, determine the majority base"""
        
    # Only try to resolve two-base ambiguities right now
    if re.match('[MRWSYK]', p.con_base):
        agreeing_bases = re.findall('[,.]', p.read_bases)
        disagreeing_bases = re.findall('[ACTGactg]', p.read_bases)
        
        # If more bases disagree with the consensus than agree, return the new base
        if len(disagreeing_bases) > len(agreeing_bases):
            majority_base = string.upper(disagreeing_bases[0])
        else:
            majority_base = p.ref_base
        
    else:
        majority_base = p.ref_base
    
    return majority_base

def parse_indel(p):
    """Parse an indel line and determine whether to correct it or not"""
    
    # print "--\n", p.raw.rstrip()
    # print "{0} support for {1}".format(p.indel_allele_one_support, p.indel_allele_one) 
    # print "{0} support for {1}".format(p.indel_allele_two_support, p.indel_allele_two) 
    # print "{0} support for something else".format(p.indel_allele_three_support) 
    
    if (p.indel_allele_one == '*') and (p.indel_allele_one_support >= p.indel_allele_two_support):
        #print "\t==> Reference has the most support - skip"
        return False
    elif (not p.indel_allele_one == '*') and (p.indel_allele_one_support > p.indel_allele_two_support):
        #print "\t==> Correct this"
        return True
    else:
        #print "\t Tie?"
        return False        

def main():
    """Correct a fasta file based on pileup-formatted SNP calls"""

    opt, args = parse_options()
    pileup_file, reference_file, output_file, edits_list = args[0], args[1], args[2], args[3]

    # Parse the header dictionary
    header_dict = build_header_dict_from_fai(reference_file)
    print "Parsed header dictionary, found {0} sequences".format(len(header_dict.keys()))
        
    # Parse the pileup file
    # Temporarily uncommented - add an option so we can use either...
    #correction_dict = parse_pileup(pileup_file, opt)
    
    # Parse a VCF file
    correction_dict = parse_VCF(pileup_file, opt)    
    print "Parsed pileup dictionary, found {0} contigs in need of correction".format(len(correction_dict.keys()))    
        
    # Parse and correct the assembly fasta file
    correct_assembly_file(header_dict, correction_dict, reference_file, output_file, edits_list)    
    print "Wrote new reference file {0}".format(output_file)
    
        
if __name__ == '__main__':
    main()

