#!/usr/bin/env python
# encoding: utf-8
"""
restore_fasta_headers.py

Created by Rod Docking on 2011-04-04.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os

from fasta_laundry import FastaRecord

def recalculate_length(G):
    """ Update contig length if necessary """

    # Note that this is ABySS-specific: otherwise we should just leave the header alone

    header_list = G.raw_header.split()
    
    try:
        seq_len = int(header_list[1])
        if not seq_len == len(G.long_seq):
            sys.stderr.write("Updating contig length for {0}: {1}->{2}\n".format(G.header_id, header_list[1], len(G.long_seq)))
            header_list[1] = str(len(G.long_seq))
            G.raw_header = " ".join(header_list)
    except (IndexError, ValueError):
        pass
    
    return G

def main():
    """ Lift-over the header from one fasta file to another, and write sequences on a single line """

    # Grab headers from the first file (usually an ABySS assembly)
    #fasta_list = []
    fasta_dict = {}
    current_header, current_seq_list = '', []    
    
    with open(sys.argv[1], 'r') as abyss_handle:
        for line in abyss_handle:
            
            if (current_seq_list != []) and (">" in line):

                F = FastaRecord(current_header, current_seq_list)
                #fasta_list.append(F) 
                fasta_dict[F.header_id] = F
                current_header, current_seq_list = '', []

            if ">" in line:
                current_header = line.rstrip()
            else:
                current_seq_list.append(line.rstrip())        

        F = FastaRecord(current_header, current_seq_list)
        #fasta_list.append(F) 
        fasta_dict[F.header_id] = F

    sys.stderr.write("Parsed first assembly, found {0} entries\n".format(len(fasta_dict.keys())))
    
    # Read in the second file, and lift over the original headers
    current_header, current_seq_list = '', []    
    with open(sys.argv[2], 'r') as scrub_handle:
        for line in scrub_handle:
            
            if (current_seq_list != []) and (">" in line):

                G = FastaRecord(current_header, current_seq_list)
                F = fasta_dict[G.header_id]
                G.raw_header = F.raw_header
                
                # Recalculate seq length
                G = recalculate_length(G)
                sys.stdout.write("{0}\n{1}\n".format(G.raw_header, G.long_seq))
                
                current_header, current_seq_list = '', []

            if ">" in line:
                current_header = line.rstrip()
            else:
                current_seq_list.append(line.rstrip())        

        G = FastaRecord(current_header, current_seq_list)
        F = fasta_dict[G.header_id]
        G.raw_header = F.raw_header
        G = recalculate_length(G)
        sys.stdout.write("{0}\n{1}\n".format(G.raw_header, G.long_seq))
                    

if __name__ == '__main__':
    main()

