#!/usr/bin/env python
# encoding: utf-8
"""
scaffold_filler.py

Created by Rod Docking on 2010-06-30.
Copyright (c) 2010 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys, os, subprocess
from lib.common_utils import build_seq_dict
from optparse import OptionParser

def find_scaffold_locations(seq_dict):
    """Parse the sequence dictionary, and find scaffolded gaps"""
    
    gap_dict = {}
    
    for key in seq_dict.iterkeys():
        seq = seq_dict[key]
        #print key, '\n', len(seq_dict[key])
        
        gap_dict[key] = []
        in_gap = False
        current_gap = [0, 0]
        
        for i in range(len(seq_dict[key])):
            
            if seq[i] == 'N' and (not in_gap):
                in_gap = True
                current_gap[0] = i
            elif seq[i] == 'N' and (in_gap):
                continue
            elif seq[i] != 'N' and (in_gap):
                in_gap = False
                current_gap[1] = i-1
                gap_dict[key].append(current_gap)
                current_gap = [0, 0]        
    
    return gap_dict            
    

def parse_options():
    """Parse options provided on the command line"""

    #print "{0} [rev{1}]".format(sys.argv[0], get_current_svn_rev(sys.argv[0]))

    usage = "usage: %prog [options] reference.fasta [mod_ref.fa]"
    parser = OptionParser(usage)

    (options, args) = parser.parse_args()
    if len(args) > 2:
        parser.error("incorrect number of arguments")
        
    # Replace arguments with full path names
    pathed_args = [os.path.abspath(arg) for arg in args]

    # Check that all files exist
    for arg in pathed_args:
        if not os.path.exists(arg):
            parser.error("{0} does not exist!".format(arg))
    
    return options, pathed_args

def extract_flanking_sequence(mod_reference, gap, coord, seq_dict):
    """Extract the original contig sequence using samtools faidx"""

    flank_start = max(coord[0]-100, 0)
    flank_stop = min(coord[1]+100, len(seq_dict[gap]))
    coord_string = "{0}:{1}-{2}".format(gap, flank_start, flank_stop)
    
    process = subprocess.Popen(
        [ 'samtools', 'faidx', mod_reference, coord_string],
        shell=False, cwd=os.getcwd(), stdout=subprocess.PIPE)
    handle = process.stdout
    contig_handle = open("replacement_sequences.fa",'a')
    for line in handle:
        contig_handle.write(line)
    handle.close()
    contig_handle.close()    
    #print mod_reference, gap, coord, flank_start, flank_stop, len(seq_dict[gap])

    return
    
def main():
    """Find scaffolds and attempt to fill them!"""

    options, args = parse_options()
    reference = args[0]
    if len(args) == 2:
        mod_reference = args[1]

    seq_dict = build_seq_dict(reference)
    
    # For each sequence, check for scaffold gaps
    gap_dict = find_scaffold_locations(seq_dict)
    num_gaps, num_contigs_with_gap, num_bases_in_gaps = 0, 0, 0
    
    # Go through the gap dictionary, and count gaps and bases in gaps
    for gap in gap_dict:
        if gap_dict[gap] == []:
            continue
    
        #print gap, gap_dict[gap]
        
        num_contigs_with_gap += 1
        for indiv_gap in gap_dict[gap]:
            num_gaps += 1
            num_bases_in_gaps += (indiv_gap[1]-indiv_gap[0]+1)
            #print "{0}\t{1}\t{2}\tscaffold".format(gap, indiv_gap[0], indiv_gap[1])
            print "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}".format(
                gap, ".", "scaffold", indiv_gap[0], indiv_gap[1], ".", ".", ".", ".")
            
    
    # Output header
    #print num_gaps, " distinct gaps in ", num_contigs_with_gap, " distinct contigs, with ", num_bases_in_gaps, " bases in gaps"
    sys.stderr.write("{0} distinct gaps in {1} distinct contigs, with {2} bases in gaps\n".format(num_gaps, num_contigs_with_gap, num_bases_in_gaps))
    
    # If a second reference is supplied, extract the sequences corresponding to the ORIGINAL gaps from the MODIFIED reference
    if len(args) == 2:
        print "Extracting fasta subsequences from modified reference ", mod_reference
        
        for gap in gap_dict:
            if gap_dict[gap] == []:
                continue
        
            for indiv_gap in gap_dict[gap]:
                extract_flanking_sequence(mod_reference, gap, indiv_gap, seq_dict)
        
        
            
    sys.exit()
                
if __name__ == '__main__':
    main()

